/*
  This file is a part of Prime Crawler,
  Copyright (C) 2009, Melikamp.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  without any warranty; compiling, interpreting, executing or merely
  reading the text of the program may result in lapses of consciousness
  and/or very being, up to and including the end of all existence and
  the Universe as we know it.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License
  along with this program (most likely, a file named COPYING).
  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <fstream>
#include <getopt.h>
#include <time.h>
#include <stdlib.h>

#include "prime-crawler.h++"

using namespace std;

/* This baby is designed to create pads and test them.  It has a lot
   of options. Dig in. */
int main(int argc, char **argv){

  /* Variables used for long option processing. */

  /* The PRNG object. */
  class PrimeCrawler * pc = NULL;
  /* The size of the binary input file in bits. */
  uli file_size;
  /* Test length is a parameter for several tests. */
  uli test_length = 42;
  /* Number of lines in the dynamically generated pad.
     -1 will be used as a negative flag value. */
  sli pad_size = -1;
  /* The seed. */
  uli seed = 0;
  /* We give them NULLs because they will be used as flags. */
  char * text_pad_file_name = NULL;
  char * binary_file_name = NULL;

  /* Automatic flags. We use ints for compatibility with struct
     option. */
  int c, option_index;
  int test_rand_flag = 0;
  int test_pc_flag = 0;
  int help_flag = 0;
  int rand_char_flag = 0;
  int rand_bin_flag = 0;
  int rand_int_flag = 0;
  int rand_string_flag = 0;
  int foo_pad_flag = 0;

  /* Bit flag bundle for various levels of reporting. */
  uli echo_flags = 0;

  /* Unless automatic flags, options need ids. The first ID,
     FLAG, is given to automatic flags. */
  enum option_name {FLAG,
		    TEXT_PAD,
		    BINARY_TO_TEXT,
		    TEST_LENGTH,
		    PAD_SIZE,
		    SEED,
		    ECHO_PAD,
		    ECHO_LINES, // Implies ECHO_PAD.
		    ECHO_FAST_LINES,
		    ECHO_STATE,
		    ECHO_ALL
  };

  struct option long_options[] = {
    /* These options set a flag automatically. */
    /* Print out usage help. */
    {"help", no_argument, &help_flag, 1},
    /* Test built-in rand by invoking it the given number of times */
    {"test-rand", no_argument, &test_rand_flag, 1},
    /* Ditto, for pc. This option will not init pc, merely check it. */
    {"test-pc", no_argument, &test_pc_flag, 1},
    /* Print a bunch of pseudo random numbers to cout, char by char. */
    {"rand-char", no_argument, &rand_char_flag, 1},
    /* Ditto, in text. */
    {"rand-int", no_argument, &rand_int_flag, 1},
    /* Again in text, but this time in binary, with lower bits on the
       left. */
    {"rand-bin", no_argument, &rand_bin_flag, 1},
    /* Use the built-in pad. */
    {"foo-pad", no_argument, &foo_pad_flag, 1},

    /* These options don't set a flag automagically. */
    /* Reads a pad from file and produces some output. */
    {"text-pad", required_argument, NULL, TEXT_PAD},
    /* Reads a binary file, converts it into a pad and prints the pad
       to cout. */
    {"binary-to-text", required_argument, NULL, BINARY_TO_TEXT},
    {"test-length", required_argument, NULL, TEST_LENGTH},
    {"pad-size", required_argument, NULL, PAD_SIZE},
    {"seed", required_argument, NULL, SEED},
    {"echo-pad", no_argument, NULL, ECHO_PAD},
    {"echo-lines", no_argument, NULL, ECHO_LINES},
    {"echo-fast-lines", no_argument, NULL, ECHO_FAST_LINES},
    {"echo-state", no_argument, NULL, ECHO_STATE},
    {"echo-all", no_argument, NULL, ECHO_ALL},
    {0, 0, 0, 0}
  };

  while(1){
    option_index = 0;
    c = getopt_long(argc, argv, "", long_options, &option_index);
    if(c == -1)
      break;
    switch(c){
    case FLAG:
      /* This option is an automatic flag. Do nothing. */
      break;
    case TEST_LENGTH:
      test_length = atol(optarg);
      break;
    case PAD_SIZE:
      pad_size = atol(optarg);
      break;
    case SEED:
      seed = atol(optarg);
      break;
    case ECHO_PAD:
      echo_flags |= PrimeCrawler::ECHO_PAD;
      break;
    case ECHO_LINES:
      echo_flags |= PrimeCrawler::ECHO_LINES | PrimeCrawler::ECHO_PAD;
      break;
    case ECHO_FAST_LINES:
      echo_flags |= PrimeCrawler::ECHO_FAST_LINES;
      break;
    case ECHO_STATE:
      echo_flags |= PrimeCrawler::ECHO_STATE;
      break;
    case ECHO_ALL:
      echo_flags = uli(0) - uli(1);
      break;
    case TEXT_PAD:
      text_pad_file_name = optarg;
      break;
    case BINARY_TO_TEXT:
      binary_file_name = optarg;
      break;
    }
  }

  /* Greet the user. */

  cerr << "Prime Crawler by Melikamp, 2009." << endl;

  /* Print out usage help? */

  if(help_flag){
    cerr << "\nprime-crawler is a command line tool for generating\n"
	 << "pseudo random numbers. See prime-crawler.pdf for the\n"
	 << "description of the algorithm and USAGE for the complete\n"
	 << "list of all command line options with verbose\n"
	 << "explanations.\n\n"
	 << " --help                   Show a short list of CL options.\n"
	 << " --text-pad <file>        Load the pad from <file>.\n"
	 << " --foo-pad                Use the default pad.\n"
	 << " --pad-size <size>        Generate a pad with <size> lines.\n"
	 << " --seed <seed>            Seed the PRNG with <seed>.\n"
	 << " --test-length <length>\n"
	 << " --rand-int               Print <length> integers.\n"
	 << " --rand-bin               Print <length> binary strings.\n"
	 << " --rand-char              Print <length> integers char by char.\n"
	 << " --binary-to-text <file>  Print <file> as a pad.\n"
	 << " --echo-pad               Print info about the loaded pad.\n"
	 << " --echo-lines             Print even more info.\n"
	 << " --echo-state             Print state info.\n"
      ;
  }

  /* Read binary file and print it out as text pad? */

  if(binary_file_name != NULL){
    ifstream foo_file;
    const uli fll = 3; // The length of the first line.

    foo_file.open(binary_file_name, ifstream::in|ifstream::binary);
    if(!foo_file.is_open()){
      cerr << "prime-crawler: Failed to open "
	   << binary_file_name << " for reading." << endl;
    }else{
      foo_file.seekg(0, ios_base::end);
      file_size = foo_file.tellg() * 8; // The files size is in bits.
      /* Determine how many lines can be extracted from file. */
      { uli n, i, j, k, bit, sum;
	char c;
	for(i=0, n=fll, sum=fll; sum <= file_size; i++)
	  sum += (n = PrimeCrawler::next_prime_after(n));
	/* There will be i lines. */
	foo_file.seekg(0);
	cout << i << endl;
	bit = 8; // Meaning that a char has to be read.
	for(k=fll; k<n; k = PrimeCrawler::next_prime_after(k)){
	  for(j=0; j<k; j++){
	    if(bit == 8){
	      foo_file.read(&c, 1);
	      bit = 0;
	    }
	    cout << (('0' + ((unsigned char)c >> bit++) & 1));
	  }
	  cout << endl;
	}
      }
      foo_file.close();
    }
  }

  /* Init from text pad file? */

  if(text_pad_file_name != NULL){
    pc = new PrimeCrawler(text_pad_file_name);
    if(!pc->is_good())
      cerr << "prime-crawler: Error while reading pad from file." << endl;
  }

  /* Use the built-in pad? */

  if(foo_pad_flag){
    if(pc != NULL){
      cerr << "prime-crawler: ignoring --foo-pad" << endl;
    } else {
      pc = new PrimeCrawler();
    }
  }

  /* Generate the pad? */

  if(pad_size >= 0){
    if(pc != NULL){
      cerr << "prime-crawler: ignoring --pad-size" << endl;
    } else {
      pc = new PrimeCrawler(pad_size);
    }
  }

  /*** Work which requires a pad. ***/

  if(pc == NULL)
    cerr << "Not using a pad." << endl;
  else {
    /* Check if the pad good.  */
    if(!pc->is_good())
      cerr << "Using an empty pad." << endl;

    /* Seed? */
    pc->seed(seed);

    /* Print info. */
    pc->echo_pad_info(cerr, echo_flags);

    /* Print out pseudo random numbers? */

    if(rand_bin_flag || rand_char_flag || rand_int_flag){
      uli i, b, x;
      unsigned char * c;

      for(i=0; i<test_length; i++){
	x = pc->rand();
	if(rand_char_flag){
	  c = (unsigned char*)(&x);
	  for(b=0; b<sizeof(uli); b++)
	    cout << *(c+b);
	}
	if(rand_bin_flag)
	  cout << PrimeCrawler::uli_to_string(x) << endl;
	if(rand_int_flag)
	  cout << x << endl;
      }
    }
  }

  /* Test speed? */

  { uli i, n, start_clock, end_clock;

    if(test_rand_flag){
      cerr << "Testing vanilla rand() " << test_length << " times." << endl;
      start_clock = clock();
      for(i=0; i<test_length; i++)
	n = rand();
      end_clock = clock();
      cerr << test_length << " calls in "
	   << (float)(end_clock - start_clock) / (float)(CLOCKS_PER_SEC)
	   << " seconds" << endl;
    }

    if(test_pc_flag){
      if(pc != NULL){
	start_clock = clock();
	cerr << "Testing pc " << test_length << " times." << endl;
	for(i=0; i<test_length; i++)
	  n = pc->rand();
	end_clock = clock();
	cerr << test_length << " calls in "
	     << (float)(end_clock - start_clock) / (float)(CLOCKS_PER_SEC)
	     << " seconds" << endl;
      }
    }
  }

  /* No parameters? */

  if(argc == 1){
    cerr << "Call with --help for a concise list of options" << endl
	 << "or see USAGE for more help." << endl;
  }

}
