/*
  This file is a part of Prime Crawler,
  Copyright (C) 2009, Melikamp.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  without any warranty; compiling, interpreting, executing or merely
  reading the text of the program may result in lapses of consciousness
  and/or very being, up to and including the end of all existence and
  the Universe as we know it.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License
  along with this program (most likely, a file named COPYING).
  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <fstream>
#include <string>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>

#include "prime-crawler.h++"

using namespace std;

/* We use this to create the default object. */
const string PrimeCrawler::foo_pad[] = {
  "011",
  "01101",
  "1110010",
  "00110010110",
  "1011010011001",
  "01010011011100011",
  "0100100111011001001",
  "11000101110000001001110",
  "00101010111000101110101001100",
  "1110110010001101010011001001110",
  "1110011000000100101011100100001011110",
  "11011110110110101100000010011110110011001",
  "1000000100101010100011011010011110110011101",
  "11011001101010000010101101110101010001011100101",
  "00111110110110101001010001110001010000110001101110110",
  "10001011001011110111101000010010101100011010101001010100100",
  "1111000111000110011001111001100100000010100111011111011110011",
  "0100100000110110111011110111000111101110111001101110010000001001101",
  "00111101001000101101011110100011101000001110000011010110110110011010011",
  "0000001010110000101110111010110000001011101111001000100100111100011100001",
  "0111011001001000011111010011110110000101110100100000001100011011101111001010000",
  "00001010101111001011100100111100101001100011100000001100111011001001101001101110110",
  "00110110011101100010000001100110000011011010111110110011000010010111100011011000001101100"
};

/************************/
/*** Private methods. ***/
/************************/

void PrimeCrawler::make_prime_lines(uli n){
  uli i, j;

  assert(is_clean());
  assert(n > 0);

  number_of_lines = n;
  line_lengths = new uli[number_of_lines];
  line_lengths[0] = 3;
  for(i=1; i<number_of_lines; i++)
    line_lengths[i] = next_prime_after(line_lengths[i-1]);
  alloc_lines();
  for(i=0; i<number_of_lines; i++){
    lines[i] = new char[line_lengths[i]];
    for(j=0; j<line_lengths[i]; j++)
      lines[i][j] = '0';
  }
}

/* Run this in the beginning of each constructor in order to achieve
   clean() state. */
void PrimeCrawler::reset(){
  /* Indicates that no memory is allocated. */
  number_of_lines = 0;

  /* Zero all pointers. */
  lines = NULL;
  fast_lines = NULL;
  line_lengths = NULL;
  state = NULL;
}

/* Free all memory. This is a safe procedure which should work even if
   memory was only partially allocated or not allocated at all. This
   is why one should always run reset() in every constructor and use
   alloc_[fast_]lines(), since they will put NULLs where they
   belong. */
void PrimeCrawler::clean(void){
  uli i;

  if(lines != NULL)
    for(i=0; i<number_of_lines; i++)
      if(lines[i] != NULL) delete [] lines[i];
  if(fast_lines != NULL)
    for(i=0; i<number_of_lines; i++)
      if(fast_lines[i] != NULL) delete [] fast_lines[i];

  if(lines != NULL) delete [] lines;
  if(fast_lines != NULL) delete [] fast_lines;
  if(line_lengths != NULL) delete [] line_lengths;
  if(state != NULL) delete [] state;

  reset();
}

/* Allocates memory for lines. It trusts that no memory was allocated
   yet. */
void PrimeCrawler::alloc_lines(void){
  uli i;
  lines = new char*[number_of_lines];
  for(i=0; i<number_of_lines; i++)
    lines[i] = NULL;
}

/* Allocates memory for fast lines. It trusts that no memory was
   allocated yet. */
void PrimeCrawler::alloc_fast_lines(void){
  uli i;
  fast_lines = new uli*[number_of_lines];
  for(i=0; i<number_of_lines; i++)
    fast_lines[i] = NULL;
}

/* This routine does a basic sanity check:

   1. The pad has at least 1 line.
   2. The length of each line is prime.
   3. Each line consists of '0's and '1's only.

   Because there are many ways in which we init the pad, it is good to
   have some assurance that there are no crazy things going on. We always
   run this just before calling make_fast_lines().
*/
bool PrimeCrawler::is_pad_good(void){
  uli i, j;

  /* 1. */
  if(number_of_lines == 0){
    cerr << "PrimeCrawler::is_pad_good: The pad is empty." << endl;
    return false;
  }

  for(i=0; i<number_of_lines; i++){
    /* 2. */
    if(!is_prime(line_lengths[i])){
      cerr << "PrimeCrawler::is_pad_good: line " << i
	   << " is of non-prime length "
	   << line_lengths[i] << "." << endl;
      return false;
    }

    /* 3. */
    for(j=0; j<line_lengths[i]; j++){
      switch(lines[i][j]){
      case '0':
      case '1':
	break;
      default:
	cerr << "PrimeCrawler::is_pad_good: "
	     << "Bad character (code " << (uli)lines[i][j]
	     << ") at lines[" << i << "][" << j
	     << "], expecting '0' or '1'." << endl;
	return false;
      }
    }
  }
  return true;
}

bool PrimeCrawler::is_clean(void){
  if(number_of_lines == 0)
    return true;
  else
    return false;
}

/* Creates fast_lines and state, based on lines. Trusts that
   is_pad_good(). */
void PrimeCrawler::make_fast_lines(void){
  uli i, j, b, li, k;
  assert(!is_clean());
  alloc_fast_lines();

  for(i=0; i<number_of_lines; i++){
    fast_lines[i] = new uli[line_lengths[i]];

    k = 0;
    for(j=0; j<line_lengths[i]; j++){
      fast_lines[i][j] = 0;
      for(b=0; b<sizeof(uli)*8; b++, k++){
	if(k==line_lengths[i])
	  k = 0;
	// I am so smart, I am so smart, S-M-R-T, S-M-R-T.
	fast_lines[i][j] |= (uli)(lines[i][k] - (uli)'0') << b;
      }
    }
  }
  state = new uli[number_of_lines];
  for(i=0; i<number_of_lines; i++)
    state[i] = 0;
}

void PrimeCrawler::init_with_foo_pad(void){
  uli i, j;

  clean();

  make_prime_lines(foo_pad_size);

  for(i=0; i<foo_pad_size; i++)
    for(j=0; j<line_lengths[i]; j++)
      lines[i][j] = foo_pad[i][j];

  if(!is_pad_good()){
    clean();
    cerr << "PrimeCrawler::init_with_foo_pad: Malformed pad." << endl;
    return;
  }

  make_fast_lines();
}

uli PrimeCrawler::init_with_text_stream(istream & pad_istream){
  uli i, j, n, k;
  string s;

  clean();

  if(!pad_istream.good()){
    cerr << "PrimeCrawler::init_with_text_stream: "
	 << "The stream cannot be read."
	 << endl;
    return 1;
  }

  /* The input stream should start with the number of lines in the
     pad.  We will keep reading until a good value is obtained. */
  n = 0;
  while(n == 0){
    if(pad_istream.eof()){
      cerr << "PrimeCrawler::init_with_text_stream: "
	   << "Read EOF, expecting number of lines."
	   << endl;
      return 1;
    }
    /* We tried doing pad_istream >> s and... We just could not figure
       out what that does. It just skips over white space like crazy
       and doesn't always set the eof bit. But it works
       sometimes. It's a bloody mess. */
    getline(pad_istream, s);
    n = atol(s.c_str());
  }
  number_of_lines = n;

  /* Reading the pad now. The line lengths are taken as they are in
     the stream. Here we start allocating the memory, so we must
     clean before returning with an error. */
  line_lengths = new uli[number_of_lines];
  alloc_lines();

  i = 0; // Indicates the current line.
  while(!pad_istream.eof()){
    getline(pad_istream, s);
    n = 0; // Now n counts good characters in the string.
    for(j=0; j<s.length(); j++)
      if(s[j] == '0' || s[j] == '1')
	n++;
    // If a string is malformed, skip it; do not advance i.
    if(n == 0) continue;

    line_lengths[i] = n;
    lines[i] = new char[line_lengths[i]];
    k = 0;
    for(j=0; j<s.length(); j++)
      if(s[j] == '0' || s[j] == '1'){
	lines[i][k] = s[j];
	k++;
      }

    // cerr << i << endl;

    if(++i == number_of_lines)
      break;
  }

  /* Did we read the correct number of lines from the stream? */
  if(i != number_of_lines){
    cerr << "PrimeCrawler::init_with_text_stream: "
	 << "Read EOF, expecting "<< number_of_lines - i << " more line(s)."
	 << endl;
    clean();
    return 1;
  }

  if(!is_pad_good()){
    cerr << "PrimeCrawler::init_with_text_stream: "
	 << "Malformed pad." << endl;
    clean();
    return 1;
  }

  make_fast_lines();

  return 0;
}

/***********************/
/*** Public methods. ***/
/***********************/

PrimeCrawler::~PrimeCrawler(){
  clean();
}

/*****************/
/* Constructors. */
/*****************/

/* Default. */
PrimeCrawler::PrimeCrawler(){
  reset();
  init_with_foo_pad();
}

/* Create an object with n lines. */
PrimeCrawler::PrimeCrawler(uli n){
  reset();
  if(n == 0)
    // This is !is_good(), i.e. trivial, so this->rand() will return 0.
    return;

  uli i, j;
  // We will use the default object to generate this pad.
  PrimeCrawler pc;
  make_prime_lines(n);
  for(i=0; i<number_of_lines; i++)
    for(j=0; j<line_lengths[i]; j++)
      lines[i][j] = ((pc.rand()%2 == 0) ? '0' : '1');

  if(!is_pad_good()){
    clean();
    cerr << "PrimeCrawler::PrimeCrawler(uli): Malformed pad."
	 << endl;
    return;
  }

  make_fast_lines();
}

/* Init from a file. See USAGE or init_with_text_stream definition for
   what the file should be like. */
PrimeCrawler::PrimeCrawler(char * file_name){
  ifstream foo_file;

  reset();

  foo_file.open(file_name, ifstream::in|ifstream::binary);
  if(!foo_file.is_open()){
    cerr << "PrimeCrawler::PrimeCrawler(char*): Failed to open "
	 << file_name << " for reading." << endl;
    return;
  }

  init_with_text_stream(foo_file);

  foo_file.close();
}

/**************************)*/
/* The most useful methods. */
/****************************/

/* This works even if the object is empty, but then it gives all
   zeroes. */
uli PrimeCrawler::rand(void){
  uli i, r;
  r = 0;
  for(i=0; i<number_of_lines; i++){
    r ^= fast_lines[i][state[i]];
    if(++(state[i]) == line_lengths[i]) state[i] = 0;
  }
  return r;
}

/* The way it is implemented, the results may be subpar for pads with
   periods at or below max uli (we are basically skipping first s long
   int outputs), but anything bigger should work just fine. */
void PrimeCrawler::seed(uli s){
  uli i, j;
  uli log_2_period = 0;

  /* Find the minimal sub-pad with period as long as max uli. */
  for(i=0; i<number_of_lines; i++)
    if((log_2_period += log2(line_lengths[i])) >= 8*sizeof(uli))
      break;

  /* We will move state markers in the first i lines. */
  for(j=0; j<i; j++)
    state[j] = s % line_lengths[j];
}

/*****************/
/* get_ methods. */
/*****************/

/* Returns the number of lines in the pad. */
uli PrimeCrawler::get_nol(void){
  return number_of_lines;
}

/* Returns the length of a line as a number of binary digits. */
uli PrimeCrawler::get_line_length(uli i){
  /* If the object is empty or the argument is out of bounds, zero is
     returned. */
  if(is_clean() || number_of_lines <= i)
    return 0;
  return line_lengths[i];
}

/* Only counts bytes used for contents of lines and fast_lines, as
   these may be expected to be much larger than number_of_lines. The
   size is in bytes. */
uli PrimeCrawler::get_footprint_lb(){
  uli i, m;
  m = 0;
  for(i=0; i<number_of_lines; i++)
    m += line_lengths[i];
  m *= (sizeof(uli) + sizeof(char));
  return m;
}

/***********************/
/* Methods that print. */
/***********************/

/* Converts an unsigned integer to string, lower bits on the left. */
string PrimeCrawler::uli_to_string(uli n){
  string s;
  uli b;
  s = "";
  for(b=0; b<8*(sizeof(uli)); b++){
    if((n & 1) == 0)
      s.append("0");
    else
      s.append("1");
    n >>= 1;
  }
  return s;
}

/* Prints various info about the pad to out. */
void PrimeCrawler::echo_pad_info(ostream & out, uli flags){
  uli zero_count, one_count;
  uli i, j;
  long double log_10_period;
  long double log_2_period;

  /* More work needed: echo basic randomness metrics, check for dupes.
   */
   log_10_period = 0;
   log_2_period = 0;
   for(i=0; i<number_of_lines; i++){
     log_10_period += log10(line_lengths[i]);
     log_2_period += log2(line_lengths[i]);
   }

   /* Here we print out info about the pad. */
   if(flags & ECHO_PAD){
     out << "Current pad:" << endl;
     out << "Line #: line length, (0 count, 1 count), line" << endl;
     for(i=0; i<number_of_lines; i++){
       out << "Line " << i << ": " << line_lengths[i] << ", ";
       zero_count = one_count = 0;
       for(j=0; j<line_lengths[i]; j++){
	 if(lines[i][j] == '0')
	   zero_count++;
	 else
	   one_count++;
       }
       out << "(" << zero_count << ", " << one_count << ") ";
       if(flags & ECHO_LINES)
	 for(j=0; j<line_lengths[i]; j++)
	   out << lines[i][j];
       out << endl;
     }
   }

   if(flags & ECHO_FAST_LINES){
     out << "Fast lines:" << endl;
     echo_fast_lines(cerr);
   }

   if(flags & ECHO_STATE){
     uli i;
     out << "State: " << endl;
     for(i=0; i<number_of_lines; i++)
       out << state[i] << " ";
     out << endl;
   }

   out << "Pad stats:" << endl;
   out << "log(10, period) = " << log_10_period
       << " and log(2, period) = " << log_2_period << "," << endl;
   out << "A lower bound on the memory footprint is "
       << get_footprint_lb() << " bytes." << endl;
}

void PrimeCrawler::echo_fast_lines(ostream & out){
  uli i, j;
  for(i=0; i<number_of_lines; i++){
    out << "Line " << i << endl;
    for(j=0; j<line_lengths[i]; j++)
      out << uli_to_string(fast_lines[i][j]) << endl;
  }
}

/*****************/
/* And the rest. */
/*****************/

bool PrimeCrawler::is_good(){
  return !is_clean();
}

bool PrimeCrawler::is_prime( uli p ){
  uli d;
  if(p == 2) return true;
  if(p%2 == 0 || p == 1) return false;
  for(d=3; d<p/2; d+=2)
    if(p%d == 0)
      return false;
  return true;
}

uli PrimeCrawler::next_prime_after( uli p ){
  uli i;
  for(i=p+1;; i++)
    if(is_prime(i))
      return i;
}
