/*
  This file is a part of Prime Crawler,
  Copyright (C) 2009, Melikamp.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  without any warranty; compiling, interpreting, executing or merely
  reading the text of the program may result in lapses of consciousness
  and/or very being, up to and including the end of all existence and
  the Universe as we know it.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License
  along with this program (most likely, a file named COPYING).
  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef PRIME_CRAWLER
#define PRIME_CRAWLER

#include <iostream>

typedef unsigned long int uli; // No other ints
typedef signed long int sli; /* Except for this one. And, to be fair,
				a normal int. But we discourage
				that. */

/* Member function descriptions are here, as far as public usage
   goes. Implementation details are with the definitions.

   In order to use this class to generate random numbers, you need to
   create an object. There is a default constructor which you can use
   to construct an object with a reasonable period and memory
   requirements.

   Alternatively, you can use the constructor which allows you to set
   the size of the internal state or even the one which reads it from
   a file.

   This class writes all diagnostics and errors on cerr. cout is
   reserved for batch processing.

   Skip to public for basic usage.
*/

class PrimeCrawler {
  /* Internally, the pad is in human-readable form.  We thought a lot
     about this and decided that once we start reading pads from
     files, we want them to be human-readable there. Moreover,
     fast_lines is already a hack designed to improve performance, so
     its implementation may have to change at some point. When and if
     that happens, no one will want to convert existing pads.

     There is no "out of memory" management yet.

     To create an object one generally wants to allocate memory for
     lines, initialize lines with '0's and '1's, allocate memory for
     fast_lines, init those (from lines), and finally init state,
     which holds positions within fast_lines. Especially in case of
     reading a pad from file, this could fail anywhere (IO fail,
     malformed input), so we need to be able to retreat: free whatever
     memory we allocated and return an object that !is_good();

     Memory management works like so: after each method an object can
     be left clean or !clean. Clean means that number_of_lines is 0,
     no memory is allocated, and all pointers are NULL. !clean means
     that number_of_lines is not 0 and some memory may have been
     allocated.

     Every constructor must start by calling reset(), which puts the
     object into the clean state.
  */

  /***********************/
  /*** Private fields. ***/
  /***********************/

  /* This is the number of logical lines in the pad. 0 iff memory is
     free. We mean it: no member function should ever return unless
     it's true. */
  uli number_of_lines;

  /*************************/
  /*** Private pointers. ***/
  /*************************/

  /* Human-readable pad is an array of "lines", which are arrays of
     chars. Each char should be '0' or '1'. */
  char ** lines;

  /* Lengths in chars. */
  uli * line_lengths;

  /* Fast lines are lines compressed to the bit level and repeated
     sizeof(long int) times. This is done so that we can xor words
     (long ints). The memory consumption is atrocious, but rand() is
     very fast. */
  uli ** fast_lines;

  /* Positions in respective fast lines. */
  uli * state;

  /************************/
  /*** Private methods. ***/
  /************************/

  /* Expects a clean object. Creates n lines, starting with line
     of length 3, sets characters to '0'. */
  void make_prime_lines(uli n);

  /* Assumes that is_pad_good() returned true after the last pad edit.
     Allocs memory and creates fast lines, as well the state: all
     zeroes. */
  void make_fast_lines();

  /* Allocates memory for lines pointers and NULLs the pointer
     array. We call this only after we set number_of_lines. */
  void alloc_lines();

  /* Ditto, for fast_lines. */
  void alloc_fast_lines();

  /* NULLs all top-level pointers and sets the number of lines to
     zero. We call this
     1. At the beginning of every constructor.
     2. At the end of clean().
  */
  void reset();

  /* Logically, either some memory is allocated or none is. */
  bool is_clean(void);

  /* Free all dynamically allocated memory. */
  void clean(void);

  /* A sanity check. Every function which changes the pad lines _must_
     run this and make sure it is true before running
     make_fast_lines(), if only to crash more gracefully. This is
     different from is_good(), which is a public object triviality
     test. */
  bool is_pad_good(void);

  /* Cleans the object and inits it with a default pad. No file
     operation, everything is built in. Returns 0 if everything is OK,
     1 if the !is_pad_good(), which almost certainly indicates that
     the code is bad. In the error case, the object comes back
     clean. */
  void init_with_foo_pad(void);

  /* Cleans the object and reads a new pad from a stream. Returns 0 if
     everything is OK.  Returns 1 if the input is malformed; when that
     happens, the object comes back clean. */
  uli init_with_text_stream(std::istream &);

  /***************************/
  /*** Static built-in pad ***/
  /***************************/

  /* The one pad that is built-in. The line lengths are primes
     starting with 3, so that it is compatible with
     make_prime_lines(). */
  static const uli foo_pad_size = 23; // Why not?
  static const std::string foo_pad[];

public:
  /* This constructor will create a default object based on
     foo_pad. This may be suitable to your serious application, or it
     may not:

     There are 23 lines.
     log(10, period) = 34.075 and log(2, period) = 113.195,
     A lower bound on the memory footprint is 8649 bytes.
  */
  PrimeCrawler();

  /* This constructor will create an object with n lines by creating a
     default object and using it to init the pad. With n=0, a clean
     (and hence !good()) object is created.
  */
  PrimeCrawler(uli n);

  /* Because pads can get big, they could sometimes be read from
     files. After calling the constructor, you gotta make sure that it
     is_good(), in case if there is no such file. */
  PrimeCrawler(char * file_name);

  /* It is sometimes a good idea to check that you have a non-trivial
     object. E.g. if you init from a file and IO fails, rand() will
     not crash but will always return 0. You shoudn't need this after
     a default constructor.
  */
  bool is_good(void);

  /* Calls clean(). */
  ~PrimeCrawler();

  /* There, a pseudo-random long unsigned int. */
  uli rand(void);

  /* Prints various info about the pad to foo_out.  See flag
     descriptions below.
  */
  void echo_pad_info(std::ostream & foo_out, uli echo_flags = 0);

  /* These are flags for echoing the pad. Large (100 and more lines)
     pads may induce a LOT of output unless the flag is zero. */

  /* Print each line as a sequence of zeroes and ones. */
  static const uli ECHO_LINES = 1;
  /* Print each fast line as a sequence of zeroes and ones. */
  static const uli ECHO_FAST_LINES = 2;
  /* Print short info about each line. */
  static const uli ECHO_PAD = 4;
  /* Print short info about each line. */
  static const uli ECHO_STATE = 8;

  /* These 2, I think, are unambiguous. */
  static uli next_prime_after( uli p );
  static bool is_prime( uli p );

  /* Modifies state. */
  void seed(uli);

  /* Returns the lower bound on memory usage, in bytes. */
  uli get_footprint_lb(void);

  /* Returns the number of lines in the pad. */
  uli get_nol(void);

  /* Returns the length of a line i as a number of binary digits.  If
     the object is empty or the argument is out of bounds, zero is
     returned. */
  uli get_line_length(uli i);

  /* Prints a number to string in binary, e.g. "00100010", low bits on
     the left. */
  static std::string uli_to_string(uli);

  /* Prints fast lines to foo_out. */
  void echo_fast_lines(std::ostream & foo_out);
};

#endif
